#!/usr/bin/env python
# -*- coding: utf-8 -*-

# http://www.st.com/resource/en/application_note/cd00221490.pdf
# http://kbeckmann.github.io/nfc-antenna-generator/
# Microchip AN710

import math
from memoize import memoize

e24_scale = math.e ** (math.log(10) / 24)

def e24_closest(value):
    e24 = int(math.log(value) / math.log(e24_scale) + .5)
    return math.e ** e24_scale

class Trace:
    m_per_oz = 34.7e-6

    resistivity = 1.72e-8 # ohm-m
    conductivity = 1 / resistivity
    temp_co = 3.862e-3 # K^-1

    def __init__(self, width, thickness = None):
        if thickness is None:
            thickness = self.m_per_oz
        self.thickness = thickness
        self.width = width

    def section_area(self):
        return self.width * self.thickness

    def section_perimeter(self):
        return 2 * (self.width + self.thickness)

    def resistance(self, length, temp = 25):
        return self.resistivity * length / self.section_area() \
               * (1 + (self.temp_co * (temp - 25)))

    def __hash__(self):
        return hash((self.width, self.thickness))

class RectAnt:
    mu0 = math.pi * 4e-7 # H / m
    ur = 1
    omega = 2 * math.pi * 13.56e6
    tq = 35.

    def __init__(self, width, height, trace, spacing, turns, layers, offset = 0.):
        self.width = width
        self.height = height
        self.trace = trace
        self.spacing = spacing
        self.turns = turns
        self.layers = layers
        self.offset = offset

    @memoize
    def inductance(self):
        imp = self.fasthenry()
        if math.isnan(imp.imag):
            raise ValueError("Inductance is NaN")
        return imp.imag / self.omega

        diam = self.trace.section_perimeter() / math.pi
        coil_thickness = self.turns * (self.spacing + self.trace.width) - self.spacing
        width = self.width - coil_thickness
        height = self.height - coil_thickness
        area = width * height
        diag = math.sqrt(width ** 2 + height ** 2)

        x1 = width * math.log(2 * area / diam / (width + diag))
        x2 = height * math.log(2 * area / diam / (height + diag))
        x3 = 2 * (width + height - diag)
        x4 = (width + height) / 4
        return self.layers * self.mu0 * self.ur \
               / math.pi * (x1 + x2 - x3 + x4) * (self.turns ** 1.8)

    @memoize
    def length(self):
        length = 0
        pitch = self.trace.width + self.spacing
        for i in range(self.turns):
            length += (self.width + self.height - pitch * i * 4) * 2
        return length * self.layers

    @memoize
    def resistance(self):
        imp = self.fasthenry()
        return imp.real

        return self.trace.resistance(self.length())

    @memoize
    def capacitance(self):
        return self.fastcap()

    @memoize
    def self_resonant_frequency(self):
        return 1 / (2 * math.pi * math.sqrt(self.inductance() * self.capacitance()))

    def __cmp__(self, other):
        c = cmp(self.inductance(), other.inductance())
        if c:
            return c
        return cmp(id(self), id(other))

    @memoize
    def base_q(self):
        return self.omega * self.inductance() / self.resistance()

    @memoize
    def damping_resistors(self):
        if self.tq * .9 < self.base_q() < self.tq * 1.1:
            return 0.
        rq = self.omega * self.inductance() / 35. - self.resistance()
        if rq <= 0:
            return 0
        return e24_closest(rq / 2)

    @memoize
    def q(self):
        return self.omega * self.inductance() \
               / (self.resistance() + 2 * self.damping_resistors())

    def is_useful(self):
        return self.tq * .9 < self.q() < self.tq * 1.1 \
               and .3e-6 < self.inductance() < 3e-6 \
               and self.self_resonant_frequency() > 25e6

    def filename(self):
        ret = "%gx%gmm_%dx%dturns_%g+%gmm" % (
            self.width * 1000, self.height * 1000,
            self.layers, self.turns,
            self.trace.width * 1000., self.spacing * 1000,)
        if not self.offset:
            return ret
        return ret + ("_%gmm" % (self.offset * 1000))

    def info(self, detailed = False):
        r = ""
        r += "%gx%g mm rectangle antenna, %g mm offset\n" % (self.width * 1000, self.height * 1000, self.offset * 1000)
        r += "%gx%g µm trace, %g mm spacing\n" % (self.trace.width * 1e6, self.trace.thickness * 1e6, self.spacing * 1000)
        r += "%d turns, %d layers, " % (self.turns, self.layers)
        r += "%g mm long\n" % (self.length() * 1e3)
        r += "R: %.3g mΩ, " % (self.resistance() * 1e3)
        r += "L: %.3g µH, " % (self.inductance() * 1e6)
        r += "C: %.3g pF\n" % (self.capacitance() * 1e12)
        r += "Q: %.3g (%.3g damped with 2x %.1g Ω)\n" % (self.base_q(), self.q(), self.damping_resistors())
        r += "Self-resonant freq: %.2f MHz\n" % (self.self_resonant_frequency() / 1e6)
        return r[:-1]

    def paint(self, painter_class, filename = None, info = False):
        assert self.layers <= 2

        p = painter_class(self)

        if self.layers == 1:
            self.paint1(p)
        elif self.layers == 2:
            self.paint2(p)
        else:
            raise NotImplemented

        if info:
            p.comment(self.info())

        return p.save(filename)

    def kicad_mod(self, filename = None):
        import painter
        return self.paint(painter.Kicad, filename, True)

    @memoize
    def fasthenry(self, filename = None):
        import painter
        return self.paint(painter.FastHenry, filename)

    @memoize
    def fastcap(self, filename = None):
        import painter
        return self.paint(painter.FastCap, filename)

    def fasthenrypdf(self, filename = None):
        import painter
        return self.paint(painter.FastHenryImage, filename)

    def fastcappdf(self, filename = None):
        import painter
        return self.paint(painter.FastCapImage, filename)

    def paint2(self, p):
        pitch = self.trace.width + self.spacing
        inner_width = self.width - (self.turns * pitch - self.spacing)
        corner_offset = (2 - 2 ** .5) * pitch
        io_offset = 2 ** .5 * pitch

        assert abs(self.offset) < inner_width / 2 - io_offset
        assert abs(self.offset) < self.width / 2 - corner_offset - io_offset

        p.reset(-self.offset + pitch, -self.height / 2 - pitch * 2, 0)

        p.pad(1)
        p.strip(-self.offset + pitch, -self.height / 2 - pitch)
        
        for t in range(self.turns):
            self.turn_paint(p,
                            self.width / 2 - pitch * t,
                            self.height / 2 - pitch * t,
                            self.offset, pitch, self.turns - t - 1, False)
        p.push()
        p.via(1)

        for t in range(self.turns - 1, -1, -1):
            self.turn_paint(p,
                            self.width / 2 - pitch * t,
                            self.height / 2 - pitch * t,
                            self.offset, pitch, self.turns - t - 1, True)
        p.strip(-self.offset - pitch, -self.height / 2 - pitch)
        p.via(0)

        p.strip(-self.offset - pitch, -self.height / 2 - pitch * 2)
        p.pad(3)

        p.pop()
        p.reset(layer = 2)
        p.strip(-self.offset, -self.height / 2 - pitch)
        p.via(0)
        p.strip(-self.offset, -self.height / 2 - pitch * 2)
        p.pad(2)

    def paint1(self, p):
        pitch = self.trace.width + self.spacing
        inner_width = self.width - (self.turns * pitch - self.spacing)
        corner_offset = (2 - 2 ** .5) * pitch
        io_offset = 2 ** .5 * pitch

        assert abs(self.offset) < inner_width / 2 - io_offset
        assert abs(self.offset) < self.width / 2 - corner_offset - io_offset

        p.reset(-self.offset + pitch, -self.height / 2 - pitch * 2, 0)

        p.pad(1)
        p.strip(-self.offset + pitch, -self.height / 2 - pitch)
        
        for t in range(self.turns):
            self.turn_paint(p,
                            self.width / 2 - pitch * t,
                            self.height / 2 - pitch * t,
                            self.offset, pitch, self.turns - t - 1, False)
        p.via(1)
        p.strip(-self.offset, -self.height / 2 - pitch)
        p.via(0)
        p.strip(-self.offset, -self.height / 2 - pitch * 2)
        p.pad(2)

    @classmethod
    def turn_paint(cls, p, x, y, ssx, pitch, nested, rev = False):
        d = pitch * (2 - 2 ** .5)
        d2 = d * nested
        if rev:
            ssx = -ssx
        positions = [
            (-ssx + (nested + 1) * d, -y - pitch),
            (-ssx + (nested + 1) * d + pitch, -y),
            (x - d2, -y),
            (x, -y + d2),
            (x, y - d2),
            (x - d2, y),
            (-x + d2, y),
            (-x, y - d2),
            (-x, -y + d2),
            (-x + d2, -y),
            (-ssx, -y),
            ]
        if rev:
            for x, y in reversed(positions):
                p.strip(-x, y)
        else:
            for x, y in positions:
                p.strip(x, y)

    def __hash__(self):
        return hash((self.width, self.height,
                     self.trace, self.spacing,
                     self.turns, self.layers))

    @classmethod
    def find_optimal(cls, width, height):
        possible = {}
        for wd in range(150, 600, 10):
            trace = Trace(wd * 1e-6)

            for sd in range(400, 900, 10):
                spacing = sd * 1e-6
                for turns in range(2, 7):
                    for layers in range(1, 3):
                        try:
                            ant = cls(width, height, trace, spacing, turns, layers, offset = -8e-3)
                            if not ant.is_useful():
                                continue
                        except:
                            print "Error while characterizing %gx%gmm %gx%gum %gum %dx%d" % (
                                width * 1000, height * 1000,
                                trace.width * 1000 * 1000, trace.thickness * 1000 * 1000,
                                spacing * 1000 * 1000,
                                turns, layers)
                            continue

                        ucrit = ant.turns, ant.layers
                        if ucrit not in possible:
                            possible[ucrit] = ant
                        else:
                            other = possible[ucrit]
                            if abs(other.q() - 35) > abs(ant.q() - 35):
                                possible[ucrit] = ant

        return [ant for k, ant in sorted(possible.items())]

def ti_rf430cl330h():
    trace = Trace(.35e-3)
    ant1 = RectAnt(34.5e-3, 39.5e-3, trace, .15e-3, 2, 1)
    ant2 = RectAnt(34.5e-3, 39.5e-3, trace, .15e-3, 3, 1)
    l = ant1.inductance() + ant2.inductance()
    r = ant1.resistance() + ant2.resistance()
    q = RectAnt.omega * l / r
    rq = e24_closest((RectAnt.omega * l / 35. - r) / 2)
    q2 = RectAnt.omega * l / (r + 2 * rq)
    print "L:", l
    print "R:", r
    print "Q:", q
    print "Rq:", rq
    print "Q':", q2

def adafruit_pn532():
    trace = Trace(.35e-3)
    ant = RectAnt(58e-3 - 4e-3, 46e-3 - 4e-3, trace, .65e-3, 4, 1)
    ant.damping_resistors = lambda : 1.5
    print ant.info()
    ant.kicad_mod()
    ant.fasthenrypdf()

def test2():
    trace = Trace(.4e-3)
    ant = RectAnt(35e-3, 35e-3, trace, .4e-3, 4, 1)
    print ant.info()

def show_optimal(w, h):
    possible = RectAnt.find_optimal(w, h)
    for i, ant in enumerate(possible):
        print ant.info()
        print

        if i >= 10:
            print len(possible) - 10, "more ommitted"
            break

def k42():
    ant = RectAnt(32e-3, 32e-3, Trace(.59e-3), .89e-3, 4, 2, offset = -6e-3)
    print ant.info()
    ant.kicad_mod()
#    ant.fasthenrypdf()
    ant.fastcappdf()

def k51():
    ant = RectAnt(32e-3, 32e-3, Trace(.58e-3), .49e-3, 5, 1, offset = -8e-3)
    print ant.info()
    ant.kicad_mod()
#    ant.fasthenrypdf()
#    ant.fastcap()
    ant.fastcappdf()
    print ant.fastcap()
    print ant.self_resonant_frequency()

def simple():
    ant = RectAnt(32e-3, 32e-3, Trace(.58e-3), .49e-3, 5, 1)
#    print ant.info()
#    ant.kicad_mod()
#    ant.fasthenrypdf()
    ant.fastcappdf()
    print ant.fastcap()
    print ant.self_resonant_frequency()

def generate_and_print(w, h, trace_w, trace_s, turns, layers, offset):
    ant = RectAnt(w / 1000., h / 1000., Trace(trace_w / 1000.), trace_s / 1000., turns, layers, offset = offset / 1000.)
    print ant.info()
    ant.kicad_mod()
#    ant.fasthenrypdf()
    ant.fastcappdf()
    print ant.fastcap()
    print ant.self_resonant_frequency()

def test_fastcap():
    class o:
        pass

    import painter
    
    ant = o()
    ant.filename = lambda: "test"
    ant.trace = o()
    ant.trace.width = 2
    ant.trace.thickness = 1
    ant.layers = 1

    p = painter.FastCap(ant)
    p.reset(1, 0, 0)
    p.strip(5, 0)
    p.strip(9, 0)
    p.strip(10, 1)
    p.strip(10, 9)
    p.strip(9, 10)
    p.strip(1, 10)
    p.strip(0, 9)
    p.strip(0, 1)
    p.strip(1, 0)

    p.save("test")

if __name__ == "__main__":
    #ti_rf430cl330h()
#    adafruit_pn532()
#    show_optimal(32e-3, 32e-3)
    #test2()
    k42()
#    for i in range(325, 400, 5):
#        generate_and_print(i/10., 32, .59, .89, 4, 2, -5)
#    simple()
#    test_fastcap()

